-- Based on LibGuild-1.0

local MAJOR_VERSION = "LibParty-1.0"
local MINOR_VERSION = tonumber(("$Revision: 81803 $"):match("%d+")) or 0

local lib, oldMinor = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local BZ

local GetNumPartyMembers = _G.GetNumPartyMembers
local select = _G.select
local next = _G.next

local new, del, clear, doAll
do
	local cache = {}
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return { ... }
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		cache[t] = true
		return nil
	end
	function clear(t)
		for k in pairs(t) do
			t[k] = nil
		end
		return nil
	end
	function doAll(func, ...)
		for i=1, select('#', ...) do
			func(select(i, ...))
		end
	end
end

local frame
if lib.frame then
	frame = lib.frame
	frame:UnregisterAllEvents()
	frame:SetScript("OnEvent", nil)
	frame:SetScript("OnUpdate", nil)
else
	frame = CreateFrame("Frame", MAJOR_VERSION .. "_Frame")
	lib.frame = frame
end

frame:SetScript("OnEvent", function(this, event, ...)
	this[event](lib, ...)
end)

local CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0")
if lib.callbacks then
	lib:UnregisterAll(lib)
else
	lib.callbacks = LibStub("CallbackHandler-1.0"):New(lib, nil, nil, "UnregisterAll")
end

frame:RegisterEvent("PARTY_MEMBERS_CHANGED")

local players = {}
local playerLevels = {}
local playerClasses = {}
local playerClassesEnglish = {}
local playersOnline = {}
local playersDead = {}
local partyLeader = nil
local numPlayersOnline = 0
local numPlayersTotal = 0
local firstScanDone = false

local pIDs = {
	"player", -- [1]
	"party1", -- [2]
	"party2", -- [3]
	"party3", -- [4]
	"party4"  -- [5]
}

local function IsInParty()
	return GetNumPartyMembers() > 0
end

local playersOnlineOld, playersOld, playerLevelsOld = {}, {}, {}
function frame:PARTY_MEMBERS_CHANGED()
	if not IsInParty() then
		doAll(clear, players, playerLevels, playerClasses, playerClassesEnglish, playersOnline, playersDead)
		numPlayersTotal = 0
		return
	end
	playersOnline, playersOnlineOld = playersOnlineOld, playersOnline -- for on/offline events
	players, playersOld = playersOld, players -- for add/remove events
	numPlayersTotal = GetNumPartyMembers()
	
	for i = 1, numPlayersTotal + 1 do
		local name, _ = UnitName(pIDs[i])
		local level = UnitLevel(pIDs[i])
		local class, englishClass = UnitClass(pIDs[i])
		local online = UnitIsConnected(pIDs[i])
		local isDead = UnitIsDeadOrGhost(pIDs[i])
		local role = UnitIsPartyLeader(pIDs[i])
		
		local add, connect

		if isDead == "" then isDead = nil end
		if name then
			players[name] = true
			playerLevels[name] = level or -1 -- note: -1 if offline
			playerClasses[name] = class or UNKNOWN
			playerClassesEnglish[name] = englishClass or UNKNOWN
			playersOnline[name] = online
			playersDead[name] = isDead

			if not playersOld[name] then
				add = true
			end
		end
		if online then
			numPlayersOnline = numPlayersOnline + 1
			if name then
				playersOnline[name] = true
				if playersOnlineOld[name] then
					playersOnlineOld[name] = nil
				elseif playersOld[name] then
					connect = true
				end
			end
		end
		if firstScanDone then
			if add then
				lib.callbacks:Fire("Added", name)
			end
			if connect then
				lib.callbacks:Fire("Connected", name)
			end
		end
	end
	for name in pairs(playersOld) do
		playersOld[name] = nil
		lib.callbacks:Fire("Removed", name)
	end
	for name in pairs(playersOnlineOld) do
		playersOnlineOld[name] = nil
		lib.callbacks:Fire("Disconnected", name)
	end
	if firstScanDone == false then
		lib.callbacks:Fire("Joined")
		firstScanDone = true
	end	
	lib.callbacks:Fire("Update")
end

function lib:GetNumOnline()
	return numPlayersOnline
end

function lib:GetNumOffline()
	return numPlayersTotal - numPlayersOnline
end

function lib:GetNumTotal()
	return numPlayersTotal
end

local playerName = UnitName("player")

function lib:HasMember(name)
	return players[name or playerName]
end

function lib:IsPlayerOnline(name)
	return playersOnline[name or playerName] or false
end

function lib:GetLevel(name)
	return playerLevels[name or playerName]
end

function lib:GetClass(name)
	return playerClasses[name or playerName], playerClassesEnglish[name or playerName]
end

function lib:GetClassColor(name)
	local class = playerClassesEnglish[name or playerName]
	if not class then return 0.8, 0.8, 0.8 end
	local c = RAID_CLASS_COLORS[class]
	return c.r, c.g, c.b
end

function lib:GetClassHexColor(name)
	local class = playerClassesEnglish[name or playerName]
	if not class then return 'cccccc' end
	local c = RAID_CLASS_COLORS[class]
	return ("%02X%02X%02X"):format(c.r*255, c.g*255, c.b*255)
end

function lib:GetIsDead(name)
	return playersDead[name or playerName]
end

function lib:GetPartyLeader()
	return partyLeader
end

local sorts; sorts = {
	NAME = function(a,b)
		return a < b
	end,
	LEVEL = function(a,b)
		if playerLevels[a] == playerLevels[b] then
			return sorts.NAME(a,b)
		else
			return playerLevels[a] < playerLevels[b]
		end
	end,
	CLASS = function(a,b)
		if playerClasses[a] == playerClasses[b] then
			return sorts.LEVEL(a,b)
		else
			return playerClasses[a] < playerClasses[b]
		end
	end,
	--[[
	ZONE = function(a,b)
		if playerZones[a] == playerZones[b] then
			return sorts.CLASS(a,b)
		else
			return playerZones[a] < playerZones[b]
		end
	end,
	RANK = function(a,b)
		if playerRank[a] == playerRank[b] then
			return sorts.CLASS(a,b)
		else
			return playerRank[a] < playerRank[b]
		end
	end,
	]]
	ONLINE = function(a,b)
		if playersOnline[a] == playersOnline[b] then
			return sorts.CLASS(a,b)
		else
			return playersOnline[a] --1 or nil, you see.
		end
	end,
}

local iter = function(t)
	local n = t.n + 1
	t.n = n
	return t[n] or del(t)
end

function lib:GetIterator(sort, includeOffline)
	local sortFunc = sorts[sort or "NAME"]
	if not sortFunc then
		error('Argument #2 must be "NAME", "LEVEL", "CLASS", "ONLINE", or nil.')
	end
	
	local tmp = new()
	for k in pairs(includeOffline and players or playersOnline) do
		tmp[#tmp+1] = k
	end
	
	table.sort(tmp, sortFunc)
	tmp.n = 0
	
	return iter, tmp, nil
end

if IsInParty() then
	frame:Show()
else
	frame:Hide()
	firstScanDone = false
	lib.callbacks:Fire("Left")
end
